package djmil.cordacheckers.gameboard;

import java.time.Duration;
import java.time.Instant;
import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.gamestate.CommitSubFlow;
import djmil.cordacheckers.gamestate.FlowResponce;
import djmil.cordacheckers.gamestate.GetFlow;
import djmil.cordacheckers.gamestate.View;
import djmil.cordacheckers.gamestate.ViewBuilder;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.common.NotaryLookup;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;

public class SurrenderFlow implements ClientStartableFlow{

     private final static Logger log = LoggerFactory.getLogger(SurrenderFlow.class);

     @CordaInject
     public JsonMarshallingService jsonMarshallingService;

     @CordaInject
     public MemberLookup memberLookup;

     @CordaInject
     public NotaryLookup notaryLookup;

     @CordaInject
     public UtxoLedgerService utxoLedgerService;

     @CordaInject
     public FlowEngine flowEngine;

     @Suspendable
     @Override
     public String call(ClientRequestBody requestBody) {
          SecureHash utxoTrxId = null;

          try {
               final GameCommand command = new GameCommand(GameCommand.Action.GAME_BOARD_SURRENDER);

               final UUID gameStateUuid = UUID.fromString(requestBody.getRequestBody());

               final StateAndRef<GameState> inputStateSar = this.flowEngine
                    .subFlow(new GetFlow(gameStateUuid));

               final GameResultState outputState = prepareGameResultState(inputStateSar);

               final UtxoSignedTransaction gameBoardSurrenderTrx = utxoLedgerService.createTransactionBuilder()
                    .addCommand(command)
                    .addInputState(inputStateSar.getRef())
                    .addOutputState(outputState)
                    .addSignatories(outputState.getParticipants())
                    .setNotary(inputStateSar.getState().getNotaryName())
                    .setTimeWindowUntil(Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
                    .toSignedTransaction();

               utxoTrxId = this.flowEngine
                    .subFlow(new CommitSubFlow(gameBoardSurrenderTrx, command.getCounterparty(inputStateSar)));

               final View gameStateView = this.flowEngine
                    .subFlow(new ViewBuilder(utxoTrxId));

               return new FlowResponce(gameStateView, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          } 
          catch (Exception e) {
               log.warn("Exception during processing " + requestBody + " request: " + e.getMessage());
               e.printStackTrace(System.out);
               return new FlowResponce(e, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          }
     }

     @Suspendable
     GameResultState prepareGameResultState(StateAndRef<GameState> gameStateSar) {
          final GameState gameState = gameStateSar.getState().getContractState();
          final GameBoardState gameBoard = (GameBoardState) gameState;

          final MemberX500Name myName = memberLookup.myInfo().getName();
          final MemberX500Name winnerName = gameBoard.getOpponentName(myName);
          
          return new GameResultState(gameBoard, winnerName);
     }

}
