package djmil.cordacheckers.gameproposal;

import static java.util.Objects.requireNonNull;

import java.time.Duration;
import java.time.Instant;
import java.util.Arrays;
import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.gamestate.CommitSubFlow;
import djmil.cordacheckers.gamestate.FlowResponce;
import djmil.cordacheckers.gamestate.View;
import djmil.cordacheckers.gamestate.ViewBuilder;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.Piece;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.common.NotaryLookup;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.membership.MemberInfo;

public class CreateFlow implements ClientStartableFlow{

     private final static Logger log = LoggerFactory.getLogger(CreateFlow.class);

     @CordaInject
     public JsonMarshallingService jsonMarshallingService;

     @CordaInject
     public MemberLookup memberLookup;

     @CordaInject
     public NotaryLookup notaryLookup;

     @CordaInject
     public UtxoLedgerService utxoLedgerService;

     @CordaInject
     public FlowEngine flowEngine;

     @Suspendable
     @Override
     public String call(ClientRequestBody requestBody) {
          SecureHash utxoTrxId = null;

          try {
               final GameCommand command = new GameCommand(GameCommand.Action.GAME_PROPOSAL_CREATE);

               final GameProposalState gameProposal = buildGameProposalStateFrom(requestBody);
               
               final UtxoSignedTransaction gameProposalCreateTrx = utxoLedgerService.createTransactionBuilder()
                    .addCommand(command)
                    .addOutputState(gameProposal)
                    .addSignatories(gameProposal.getParticipants())
                    .setNotary(notaryLookup.getNotaryServices().iterator().next().getName())
                    .setTimeWindowUntil(Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
                    .toSignedTransaction();

               utxoTrxId = this.flowEngine
                    .subFlow(new CommitSubFlow(gameProposalCreateTrx, command.getCounterparty(gameProposal)));

               final View gameView = this.flowEngine
                    .subFlow(new ViewBuilder(utxoTrxId));

               return new FlowResponce(gameView, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          } 
          catch (Exception e) {
               log.warn("Exception during processing " + requestBody + " request: " + e.getMessage());
               return new FlowResponce(e, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          }
     }

     @Suspendable
     private GameProposalState buildGameProposalStateFrom(ClientRequestBody requestBody) {
          final CreateFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, CreateFlowArgs.class);

          final Piece.Color opponentColor = Piece.Color.valueOf(args.opponentColor);
          if (opponentColor == null) {
               throw new RuntimeException("Allowed values for opponentColor are: "
                    + Piece.Color.WHITE.name() +", "  + Piece.Color.BLACK.name() 
                    + ". Actual value was: " + args.opponentColor);
          }

          final MemberInfo myInfo = memberLookup.myInfo();
          final MemberInfo opponentInfo = requireNonNull(
               memberLookup.lookup(MemberX500Name.parse(args.opponentName)),
               "MemberLookup can't find opponentName specified in flow arguments: " + args.opponentName
          );

          final MemberInfo whitePlayerInfo = opponentColor == Piece.Color.WHITE ? opponentInfo : myInfo;
          final MemberInfo blackPlayerInfo = opponentColor == Piece.Color.BLACK ? opponentInfo : myInfo;

          return new GameProposalState(
               whitePlayerInfo.getName(),
               blackPlayerInfo.getName(),
               myInfo.getName(), // <<--- GameProposal issuer
               UUID.randomUUID(),
               args.message,
               Arrays.asList(myInfo.getLedgerKeys().get(0), opponentInfo.getLedgerKeys().get(0))
          );
     }

}
