package djmil.cordacheckers.gamestate;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatedBy;
import net.corda.v5.application.flows.ResponderFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionValidator;

@InitiatedBy(protocol = "gamestate-commit")
public class CommitSubFlowResponder implements ResponderFlow {
    
    private final static Logger log = LoggerFactory.getLogger(CommitSubFlowResponder.class);

    @CordaInject
    public MemberLookup memberLookup;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;
    
    @Suspendable
    @Override
    public void call(FlowSession session) {
        UtxoTransactionValidator txValidator = trxToValidate -> {
            try {
                final GameCommand gameCommand = getSingleCommand(trxToValidate, GameCommand.class);
                final GameState gameState = getActualGameStateFromTransaction(trxToValidate, gameCommand);

                checkParticipants(session, gameCommand, gameState);

                /*
                 * Other checks / actions ?
                 */

                log.info("Approval for " + trxToValidate.getId());
            } catch (Exception e) {
                log.warn("Validation has failed. " +e.getMessage());
                // We have to re-trow an exception to indicate validation failure
                throw new CordaRuntimeException(e.getClass().getCanonicalName(), e.getMessage(), e.getCause());
            }
        };

        try {
            this.utxoLedgerService
                .receiveFinality(session, txValidator);
        } catch (Exception e) {
            // Log Corda's specific message about finality validation failure
            log.warn(e.getMessage()); 
        }
    }

    /**
     * 
     * @param gameStateTransaction an utxo ledger transaction, that involves GameState
     * @param command
     * @return the most recent (from perspective of session participants validation) GameState for a given transaction
     * 
     * @see djmil.cordacheckers.gamestate.ViewBuilder#getLatestGameStateFromTransaction(UtxoLedgerTransaction, GameCommand)
     */
    @Suspendable
    GameState getActualGameStateFromTransaction(UtxoLedgerTransaction gameStateTransaction, GameCommand command) {
        switch (command.action) {
            case GAME_PROPOSAL_CREATE:
                return getSingleOutputState(gameStateTransaction, GameProposalState.class);
            
            case GAME_PROPOSAL_ACCEPT:
            case GAME_PROPOSAL_REJECT:
            case GAME_PROPOSAL_CANCEL:
                return getSingleInputState(gameStateTransaction, GameProposalState.class);

            case GAME_BOARD_SURRENDER:
                return getSingleInputState(gameStateTransaction, GameBoardState.class);

            case GAME_BOARD_MOVE:
                return getSingleOutputState(gameStateTransaction, GameBoardState.class);

            case GAME_RESULT_CREATE:
                return getSingleOutputState(gameStateTransaction, GameResultState.class);
        }

        throw new GameCommand.ActionException();
    }

    @Suspendable
    void checkParticipants(FlowSession session, GameCommand gameCommand, GameState gameState) throws ParticipantException {
        final var myName = memberLookup.myInfo().getName();
        final var opponentName = gameState.getOpponentName(myName); // throws NotInvolved
        final var conterpartyName = session.getCounterparty();
        final var actorName = gameCommand.getInitiator(gameState);

        if (conterpartyName.compareTo(opponentName) != 0)
            throw new ParticipantException("Counterparty", conterpartyName, opponentName);

        if (actorName.compareTo(conterpartyName) != 0)
           throw new ParticipantException("Actor", conterpartyName, actorName);
    }

    public static class ParticipantException extends Exception {
        public ParticipantException(String role, MemberX500Name expected, MemberX500Name actual) {
            super("Bad participants. " +role +" role: expected '" +expected +"', actual '" +actual +"'");
        }
    }

}
